#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 S. Daniel Francis <francis@sugarlabs.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.

import logging
logger = logging.getLogger('option')

import gobject
import gtk

from item import Item


class ShortContentItem(Item):

    def __init__(self, parent=None, stock_id=None, important=False):
        Item.__init__(self, stock_id, important)
        self.content = gtk.EventBox()
        self.parent = parent
        self.separator = None

    def get_tool_item(self):
        self.toolitem = gtk.ToolItem()
        self.toolitem.add(self.content)
        self.setup_tooltip()
        return self.toolitem

    def do_activate(self):
        if self.tooltip:
            title = self.tooltip
        else:
            title = gtk.stock_lookup(self.stock_id)[1].replace('_', '')
        window = gtk.Dialog(title, self.parent)
        window.set_modal(False)
        window.set_decorated(True)
        window.set_has_separator(False)
        window.set_border_width(10)
        if self.toolitem:
            self.toolitem.remove(self.content)
            if self.separator:
                self.separator.hide()
        window.vbox.pack_start(self.content)
        self.content.show()
        window.connect('delete-event', self.destroy_window)
        window.show()

    def destroy_window(self, window, event):
        window.vbox.remove(self.content)
        window.destroy()
        if self.toolitem:
            self.toolitem.add(self.content)
            if self.separator:
                self.separator.show()
